from __future__ import annotations
from functools import partial
from itertools import zip_longest

import torch
from torch import nn, cat
from torch.nn import Module, ModuleList, Identity

from x_mlps_pytorch.norms import RMSNorm, LayerNorm
from x_mlps_pytorch.normed_mlp import create_mlp as create_normed_mlp

# functions

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def divisible_by(num, den):
    return (num % den) == 0

# main class

class ResidualNormedMLP(Module):
    # proposed by Wang et al. https://arxiv.org/abs/2503.14858 for scaling goal conditioned CRL

    def __init__(
        self,
        dim,
        depth = 32,
        residual_every = 4, # they used every 4
        dim_in = None,
        dim_out = None,
        activation = nn.SiLU(), # apparently silu was very important - nice ablation in paper
        bias = True,
        norm_fn: Module | None = None,
        use_rmsnorm = False,
        final_norm = True,
        skip_to_output = False,     # auto-compression network
        keel_post_ln = False,       # use the keel post-norm architecture proposed by Chen et al. from Bytedance - https://arxiv.org/abs/2601.19895
        keel_residual_scale = None  # defaults to number of blocks, for scaling the residual besides the first
    ):
        super().__init__()
        assert not (keel_post_ln and skip_to_output)

        assert divisible_by(depth, residual_every), '`depth` must be divisible by `residual_every`'

        # proj in and out

        self.proj_in = nn.Linear(dim_in, dim) if exists(dim_in) else Identity()
        self.proj_out = nn.Linear(dim, dim_out) if exists(dim_out) else Identity()

        # layers

        layers = []

        # norm type

        if not exists(norm_fn):
            norm_fn = RMSNorm if use_rmsnorm else LayerNorm

        # layers

        num_blocks = depth // residual_every

        for _ in range(num_blocks):

            block = create_normed_mlp(
                dim = dim,
                depth = residual_every,
                norm_fn = norm_fn,
                activation = activation,
                bias = bias,
                activate_last = True
            )

            layers.append(block)

        self.layers = ModuleList(layers)

        self.post_norms = ModuleList([norm_fn(dim) for _ in range(num_blocks - 1)]) if keel_post_ln else None

        self.keel_residual_scale = default(keel_residual_scale, num_blocks)

        self.skip_to_output = skip_to_output

        # final norm if not using keel post norm

        self.final_norm = norm_fn(dim) if final_norm and not keel_post_ln else Identity()

    def forward(
        self,
        x
    ):
        skip_to_output = self.skip_to_output

        if isinstance(x, (list, tuple)):
            x = cat(x, dim = -1)

        x = self.proj_in(x)

        layer_outs = []

        for layer_index, layer in enumerate(self.layers):
            first_layer = layer_index == 0

            residual = x

            out = layer(x)

            # skip to output - auto compression paper

            if skip_to_output:
                x = out
                layer_outs.append(out)
                continue

            # handle residual with maybe post norm with keel stabilisation techniques

            has_post_norms = exists(self.post_norms)

            if first_layer or not has_post_norms:
                x = out + residual
                continue

            # handle post norm

            post_norm = self.post_norms[layer_index - 1]

            residual = residual * self.keel_residual_scale

            x = out + residual

            x = post_norm(x)

        # Dorovatas et al. https://openreview.net/forum?id=eIDa6pd9iQ

        if skip_to_output:
            x = sum(layer_outs)

        x = self.final_norm(x)

        return self.proj_out(x)



import torch
from torch.nn import Module

# gradient dropout

class GradientDropout(Module):
    def __init__(
        self,
        prob
    ):
        super().__init__()
        assert 0. <= prob < 1.

        self.prob = prob
        self.keep_prob = 1. - prob
        self.scale = self.keep_prob ** -1

    def forward(self, t):

        if not t.requires_grad or not self.training:
            return t

        keep = torch.rand_like(t) < self.keep_prob
        grad_mask = keep.float() * self.scale

        # this trick was initially employed in a sparse memory update paper from Lin et al.
        # refashioned as a way to do easy gradient dropout

        return t * grad_mask + t.detach() * (1. - grad_mask)

# quick test

if __name__ == '__main__':
    grad_dropout = GradientDropout(0.5)

    t = torch.randn(50, requires_grad = True)

    out = grad_dropout(t)
    out.sum().backward()

    assert torch.allclose(t, out)

    print(t.grad) # about half of gradients dropped



import pytest
import torch
param = pytest.mark.parametrize

def identity(t):
    return t

def to_half(t):
    return t.half()

@param('low_rank', (None, 2))
@param('test_tensor_noise_scale', (False, True))
@param('float16', (False, True))
def test_noisable(
    low_rank,
    test_tensor_noise_scale,
    float16
):
    from torch import nn
    from random import randrange

    from x_mlps_pytorch.noisable import Noisable

    # first define module, wrap with Noisable, make sure forward is the same

    linear = nn.Linear(32, 64)

    noisable_linear = Noisable(linear, low_rank = low_rank)

    x = torch.randn(3, 32)

    maybe_to_float16 = to_half if float16 else identity

    x = maybe_to_float16(x)
    linear = maybe_to_float16(linear)

    out = linear(x)
    not_noised_out = noisable_linear(x)

    assert out.shape == (3, 64)
    assert torch.allclose(out, not_noised_out)

    # make sure can noise by passing in {param_name: noise}

    noised_out = noisable_linear(x, noise_for_params = dict(weight = maybe_to_float16(torch.randn(64, 32))))

    assert not torch.allclose(out, noised_out)

    # make sure can noise by passing in {param_name: seed}

    seed = randrange(int(1e7))
    noised_out = noisable_linear(x, noise_for_params = dict(weight = seed), noise_scale_for_params = dict(weight = maybe_to_float16(torch.ones((64, 32)))))

    assert not torch.allclose(out, noised_out)

    noised_out_again = noisable_linear(x, noise_for_params = dict(weight = seed))

    # able to temporarily noise the weights of a network, for rolling out against environment in memory efficient manner

    with noisable_linear.temp_add_noise_(dict(weight = (seed, 1e-4))):
        assert not torch.allclose(out, noisable_linear(x))

    assert torch.allclose(out, noisable_linear(x), atol = 1e-2 if float16 else 1e-6)



import torch
from torch import nn
from torch.nn import Module
import torch.nn.functional as F

# weight decay friendly norms
# Ohad Rubin - https://medium.com/@ohadrubin/exploring-weight-decay-in-layer-normalization-challenges-and-a-reparameterization-solution-ad4d12c24950

class RMSNorm(Module):
    def __init__(
        self,
        dim,
        eps = 1e-5
    ):
        super().__init__()
        self.gamma = nn.Parameter(torch.zeros(dim))
        self.eps = eps

    def forward(self, x):
        return F.rms_norm(x, x.shape[-1:], self.gamma + 1, eps = self.eps)

class LayerNorm(Module):
    def __init__(
        self,
        dim,
        eps = 1e-5
    ):
        super().__init__()
        self.gamma = nn.Parameter(torch.zeros(dim))
        self.eps = eps

    def forward(self, x):
        return F.layer_norm(x, x.shape[-1:], self.gamma + 1, eps = self.eps)



import pytest

import torch
from torch import cat
from torch.nn import Module

def test_critic():
    from x_mlps_pytorch.ensemble import Ensemble
    from x_mlps_pytorch.normed_mlp import MLP

    critic = MLP(10, 5, 1)
    state = torch.randn(2, 10)

    assert critic(state).shape == (2, 1)
    critics = Ensemble(critic, 10)

    assert critics(state).shape == (10, 2, 1)

    subset_ids = torch.tensor([0, 3, 5])
    assert critics(state, ids = subset_ids).shape == (3, 2, 1)

    subset_ids = torch.tensor([0, 0])
    assert critics(state, ids = subset_ids, each_batch_sample = True).shape == (2, 1)

    critic = critics.get_one(2)
    assert critic(state).shape == (2, 1)

    assert critics.forward_one(state, id = 0).shape == (2, 1)

    critic = critics.get_one([2, 3], weights = [0.1, 0.2])
    assert critic(state).shape == (2, 1)

def test_critic_multiple_args():
    from x_mlps_pytorch.ensemble import Ensemble
    from x_mlps_pytorch.normed_mlp import MLP

    class Critic(Module):
        def __init__(self, net):
            super().__init__()
            self.net = net

        def forward(self, states, actions):
            return self.net(torch.cat((states, actions), dim = -1))

    critic = Critic(MLP(12, 5, 1))

    state = torch.randn(2, 10)
    action = torch.randn(2, 2)

    assert critic(state, action).shape == (2, 1)
    critics = Ensemble(critic, 10)

    assert critics(state, action).shape == (10, 2, 1)

    subset_ids = torch.tensor([0, 3, 5])
    assert critics(state, action, ids = subset_ids).shape == (3, 2, 1)

    critic = critics.get_one(2)
    assert critic(state, action).shape == (2, 1)

    critic = critics.get_one([2, 3], weights = [0.1, 0.2])
    assert critic(state, action).shape == (2, 1)



import torch
from torch import nn, cat
from torch.nn import Linear, Module, ModuleList

from einops import repeat

# functions

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

# main class

class MLP(Module):
    def __init__(
        self,
        *dims,
        dim_latent,
        latent_mlp = False,
        latent_mlp_dim_hidden = None,
        condition_hiddens = False,
        activation = nn.ReLU(),
        bias = True,
        activate_last = False,
    ):
        super().__init__()
        assert len(dims) > 1, f'must have more than 1 layer'

        layers = []

        # concatenative conditioning

        first_dim, *rest_dims = dims

        latent_mlp_dim_hidden = default(latent_mlp_dim_hidden, first_dim * 2)

        # maybe latent mlp

        dim_cond = dim_latent

        self.to_latent_cond = nn.Identity()

        if latent_mlp:
            self.to_latent_cond = nn.Sequential(
                Linear(dim_latent, latent_mlp_dim_hidden),
                nn.SiLU(),
            )

            dim_cond = latent_mlp_dim_hidden

        # dimensions again

        first_dim_with_latent_cond = first_dim + dim_cond

        dims = (first_dim_with_latent_cond, *rest_dims)

        # input output dimension pairs

        dim_in_out = tuple(zip(dims[:-1], dims[1:]))

        # layers

        for i, (dim_in, dim_out) in enumerate(dim_in_out, start = 1):
            is_first = i == 0
            is_last = i == len(dim_in_out)

            layer = nn.Linear(dim_in, dim_out, bias = bias)

            # if not last, add an activation after each linear layer

            if not is_last or activate_last:
                layer = nn.Sequential(layer, activation)

            # maybe additional layer cond

            latent_to_layer_cond = None

            if latent_to_layer_cond and not is_first:
                latent_to_layer_cond = Linear(dim_cond, dim_in)

            layers.append(ModuleList([
                layer,
                latent_to_layer_cond
            ]))

        self.layers = ModuleList(layers)

    def forward(
        self,
        x,
        latent
    ):
        if isinstance(x, (list, tuple)):
            x = cat(x, dim = -1)

        cond = self.to_latent_cond(latent)

        if cond.ndim == 2 and x.ndim == 3:
            cond = repeat(cond, 'b d -> b n d', n = x.shape[1])
            x = cat((cond, x), dim = -1)

        for layer, maybe_cond in self.layers:

            # maybe hadamard conditioning with latent

            if exists(maybe_cond):
                x = x * maybe_cond(cond)

            x = layer(x)

        return x

# factory function

def create_mlp(
    dim,
    depth,
    *,
    dim_in = None,
    dim_out = None,
    **mlp_kwargs
):
    dims = (dim,) * (depth + 1)

    if exists(dim_in):
        dims = (dim_in, *dims)

    if exists(dim_out):
        dims = (*dims, dim_out)

    return MLP(*dims, **mlp_kwargs)
